package fr.llibre.ml2;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Collections;

/**
 *  Classe Etoile ne mémorisant que les éléments constants de l'étoile,
 *  comme son ascension droite à J2000 dans J2000, etc.
 *  tirées du BrightStar Catalog de 9110 étoiles
 *  Les éléments variables : ascension droie apparente, etc...
 *  ne sont pas mémorisés. ils sont accessibles par des méthodes
 *  "propres à l'étoile" qui les renvoient par le biais de tableaux de doubles.
 *   
 *  La classe offre aussi un certain nombre de méthodes statiques par exemple
 *  pour lire le fichiers des étoiles du Bright Star Catalog et renvoyer le
 *  tableaux correspondant.
 * 
 * @author llibre
 *
 */
public class Etoile {
	public String id; // Nom composé fichier 9110Stars ou autre identifiant d'objet céleste
//	public String bsVstarId; // Identificateur flou	fichier 9110Stars
	public double mag; // Magnitude
	public int bsNumber; // Numero dans le fichier 9110Stars
	public double ad0_hd; // Ascension droite heures décimales /J2000
	public double de0_dd; // Déclinaison degrés décimaaux /J2000
	public String tipe;	// nature de l'objet
	public double vA_sda; // Vitesse Asc. Dr. en sec de degré / an
	public double vD_sda; // Vitesse Dec.     en sec de degré / an  
//	public double vitMod_sda; // Module de cette vitesse
	public double work; // valeur à tout faire, par ex. dist à une direction
	public boolean flCoher; //  false si données incohérentes
    public double uCS[] = new double[3]; // Pour les graphiques

	/**
	 * 	   Calcul de diverses informations sur l'étoile et génération d'un tableau
	 *     de doubles et d'un string contenant ces infos.
	 *     
	 * result[0 et 1]  : Coord. équat. à J2000 / J2000 (heures et degrés)
	 * result[2 et 3]  : Coord. équat. à à la date / rep date (heures et degrés)
	 * result[4]       : Angle horaire (heures)
	 * result[5 et 6]  : Coord. horizontales, azimut et hauteur (degrés)
	 * result[7]       : Angle de réfraction à ajouter (degrés)
	 * result[8]       : Date du passage au méridien (transit) en jour julien
	 * result[9]       : Hauteur (degrés) au lors du transit
	 * result[10]      : Date du lever en jour julien (-1 si inexistant)
	 * result[11]      : Azimut au lever (degrés)
	 * result[12]      : Date du coucher en jour julien (-1 si inexistant)
	 * result[13]      : Azimut au coucher (degrés)
	 * 
	 * @param ju       : Date jour julien UT
	 * @param longi_dg : Longitude du lieu (degrés)
	 * @param lati_dg  : Latitude du lieu (degrés)
	 * @param result   : Tableau pour ranger les résultats (si non nul)
	 * @param sexaDec  : -1 : String result en décimal, 0 : neant , +1 en sexagésimal
	 * @param eco      : true => utilisation des paramètres variant lentement
	 * @return         : String contenant toutes les infos.
	 */
	public String infos(double ju, double longi_dg, double lati_dg, double result[], int sexaDec, boolean eco)
	{				
		double tde_rd[] = {0};
		double alfa_rd =  ju2equa_rd(ju, tde_rd, eco);
		double decli_rd = tde_rd[0];
		double alfa_dg = Math.toDegrees(alfa_rd);
		double decli_dg = Math.toDegrees(decli_rd);
		double lati_rd = Math.toRadians(lati_dg);
		double cosLa = Math.cos(lati_rd);
		double sinLa = Math.sin(lati_rd);
		double ttopo[] = new double[3];
		double hourAng_rd = Astro.equa2topo_rd(ju, alfa_rd, decli_rd, Math.toRadians(longi_dg),
				lati_rd, ttopo, eco);
		double hourAng_dg = Math.toDegrees(hourAng_rd);
		double azi_rd = ttopo[0];
		double azi_dg = Math.toDegrees(azi_rd);
		double haut_rd = ttopo[1];
		double haut_dg = Math.toDegrees(haut_rd);
		double hautCor_dg = ttopo[2];

		if (result != null)
		{
			result[0] = ad0_hd;
			result[1] = de0_dd;
			result[2] = alfa_dg/15.;
			result[3] = decli_dg;
			result[4] = hourAng_dg/15.;
			result[5] = azi_dg;
			result[6] = haut_dg;
			result[7] = hautCor_dg;
		}
		DecimalFormatSymbols dfs = new DecimalFormatSymbols(); dfs.setDecimalSeparator('.');
		DecimalFormat df4 = new DecimalFormat("0.0000", dfs);
		String str = "";
		if (sexaDec > 0)
		{
			str += "C. équat./J2000 : " + Utils.deg2hms(ad0_hd*15.)
					+ " | " + Utils.deg2dms2(de0_dd) + "\n";
			str += "C. équat / Date : " + Utils.deg2hms(alfa_dg)
					+ " | " + Utils.deg2dms2(decli_dg) + "\n";
			str += "C. hor. app.: " + Utils.deg2hms(hourAng_dg) + " (" + Utils.deg2dms(hourAng_dg)
					+ ") | " + Utils.deg2dms2(decli_dg) + "\n";
			str += "C. horizontales : " + Utils.deg2dms2(azi_dg)
					+ " | " + Utils.deg2dms2(haut_dg) + "\n";
			str += "Hauteur corrigée: " + Utils.deg2dms2(hautCor_dg) + "\n";
		}
		else if (sexaDec < 0)
		{
			str += "C. équat./J2000 : " + df4.format(ad0_hd)
					+ "h | " + df4.format(de0_dd) + "°\n";
			str += "C. équat / Date : " + df4.format(alfa_dg/15.)
					+ "h | " + df4.format(decli_dg) + "°\n";
			str += "C. hor. app.: " + df4.format(hourAng_dg/15.) + "h (" + df4.format(hourAng_dg)
					+ "°) | " + df4.format(decli_dg) + "°\n";
			str += "C. horizontales : " + df4.format(azi_dg)
					+ "° | " + df4.format(haut_dg) + "°\n";
			str += "Hauteur corrigée: " + df4.format(hautCor_dg) + "°\n";
		}
		// CALCUL TRANSIT
		// Hauteur de l'astre au début de sa visibilité et instant proche du transit
		double h0_dg = -0.5667; // Refraction au niveau de l'horizon
		double jref = Math.floor(ju + 0.5)+0.5; // minuit local de ce jour
		jref -=  longi_dg/360.; // date UT pour minuit local
		double sinh0 = Math.sin(Math.toRadians(h0_dg));
		// Constantes ou données initiales
		double omegaT_dgj = 360.98564736629; // vrot_terre degrés/jour
		// Ascension droite méridien local à jref
		double	Aref_dg = Astro.jul2agst_dg(jref,eco) + longi_dg;
		double deltaAsc_dg = Utils.realmod2(alfa_dg - Aref_dg, 360.);
		double jtran = jref + deltaAsc_dg/omegaT_dgj;
		double htran_dg = 90 - Math.abs(lati_dg - decli_dg);
		if (result != null)
		{
			result[8] = jtran;
			result[9] = htran_dg;
		}		
		if (sexaDec > 0)
		{
			str += "Date transit    : " + Utils.jul2str1(jtran) + " UT\n";
			str += "Hauteur transit : " + Utils.deg2dms2(htran_dg) + "\n";
		}
		if (sexaDec < 0)
		{
			str += "Date transit    : " + Utils.jul2str1(jtran) + " UT\n";
			str += "Hauteur transit : " + df4.format(htran_dg) + "°\n";
		}
		double sinDec = Math.sin(decli_rd);
		double cosDec = Math.cos(decli_rd);
		double cosH = (sinh0 - sinLa*sinDec)/(cosLa*cosDec);
		if (Math.abs(cosH) <= 1.)
		{
			double H_rd = Utils.acos(cosH);
			double H_dg = Math.toDegrees(H_rd);
			double jlev = jref + (deltaAsc_dg - H_dg)/omegaT_dgj;
			double jpon = jref + (deltaAsc_dg + H_dg)/omegaT_dgj;
			double azipon_rd = Math.atan2(Math.sin(H_rd), sinLa*cosH - cosLa*sinDec/cosDec);
			double azipon_dg = Math.toDegrees(azipon_rd);
			double azilev_dg = -azipon_dg; 
			if (result != null)
			{
				result[10] = jlev;
				result[11] = azilev_dg;
				result[12] = jpon;
				result[13] = azipon_dg;
			}
			if (sexaDec > 0)
			{
				str += "Date au lever   : " + Utils.jul2str1(jlev) + " UT\n";
				str += "Azi/Sud lever   : " + Utils.deg2dms2(azilev_dg) + "\n";
				str += "Date au coucher : " + Utils.jul2str1(jpon) + " UT\n";
				str += "Azi/Sud coucher : " + Utils.deg2dms2(azipon_dg) + "\n";
			}		
			if (sexaDec < 0)
			{
				str += "Date au lever   : " + Utils.jul2str1(jlev) + " UT\n";
				str += "Azi/Sud lever   : " + df4.format(azilev_dg) + "°\n";
				str += "Date au coucher : " + Utils.jul2str1(jpon) + " UT\n";
				str += "Azi/Sud coucher : " + df4.format(azipon_dg) + "°\n";
			}		
		}
		else 
		{
			if (result != null)
			{
				result[10] = -1;
				result[11] = 0.;
				result[12] = -1;
				result[13] = 0.;
			}
			if (sexaDec != 0)
			{
				if (htran_dg < h0_dg) str += "Toujours sous l'horizon !";
				else str += "Toujours au-dessus de l'horizon !"; 
			}
		}

		return str;
	}

    /**
     * Mise à jour des ccsinus directeurs de la direction des étoiles d'un tableau
     * dans le repère équatorial vrai de la date
     * Suppose l'appel préalable de Astro.update_eco(ju) et Astro.updateMat();
     * @param tabStars
     * @param ju
     */
	public static void updateAlluCS(Etoile[]tabStars, double ju)
	{
        double t = (ju - 2451545)/365.25;
        for (int i = 0; i < tabStars.length; i++)
		{
			Etoile e = tabStars[i];
			double ad0_rd = Math.toRadians(e.ad0_hd*15.);
			double de0_rd = Math.toRadians(e.de0_dd);
			double vA_rda = Math.toRadians(e.vA_sda/3600.);
			double vD_rda = Math.toRadians(e.vD_sda/3600.);
			double adt_rd = ad0_rd + vA_rda*t;
			double dec_rd = de0_rd + vD_rda*t;
			double cosDec = Math.cos(dec_rd);
            e.uCS[0] = Math.cos(adt_rd)*cosDec;
            e.uCS[1] = Math.sin(adt_rd)*cosDec;
            e.uCS[2] = Math.sin(dec_rd);
            R3.madvec(e.uCS, Astro.m3_CCv0, e.uCS);
		}
	}


	/**
	 *   Calcul des coord. équatoriales apparentes dans le repère de la date
	 *   
	 * @param ju     : jour julien UT.
	 * @param tde_rd : référence pour le renvoi de la déclinaison
	 * @param eco    : true => 
	 * @return
	 */
	public double ju2equa_rd(double ju, double tde_rd[], boolean eco)
	{
		double ad0_rd = Math.toRadians(ad0_hd*15.);
		double de0_rd = Math.toRadians(de0_dd);
		double vA_rda = Math.toRadians(vA_sda/3600.);
		double vD_rda = Math.toRadians(vD_sda/3600.);
		double t = (ju - 2451545)/365.25;
		double adt_rd = ad0_rd + vA_rda*t;
		double det_rd = de0_rd + vD_rda*t;
		if (eco)return  Astro.equaJ0toApp_rd_eco (ju, adt_rd, det_rd, tde_rd);
		return  Astro.equaJ0toApp_rd_std (ju, adt_rd, det_rd, tde_rd);
	}

	/**
	 *  Classe utilisée pour faire des listes d'éléments proches.
	 *  Un élément ne contient que l'index d'un autre élément dans un
	 *  tableau et la distance de cet élément à l'étoile this.
	 *  Ne contient donc que l'index et la distance et le comparateur
	 *  de distance.
	 *  
	 * @author llibre
	 *
	 */
	public static class Proxy implements Comparable<Proxy> {
		double dist; int n;
		Proxy(double d, int i){dist = d; n = i;}
		@Override
		public int compareTo(Proxy arg0) {
			double r = dist - arg0.dist;
			if (r == 0.) return 0;
			else {if (r > 0.) return 1; else return -1; }
		}		
	}
	
	/**
	 *   Ecart entre l'étoile this et une direction donnée dans J0, mais
	 *   à une date ju.
	 * @param u : cos.directeurs de la direction dans J0
	 * @param anju     : ecart à J0 en année julienne cad (ju - 2451545)/365.25	 
	 * @return         : Ecart en radians entre la direction donnée et l'étoile
	 */
	public double distanceTo(double u[], double anju)
	{		
		double ds_rd = Math.toRadians(de0_dd + vD_sda*anju/3600.);
		double as_rd = Math.toRadians(ad0_hd*15. + vA_sda*anju/3600.);
		double pscar = (u[0]* Math.cos(as_rd) + u[1]* Math.sin(as_rd))* Math.cos(ds_rd)
			+ u[2]* Math.sin(ds_rd);
		return Utils.acos(pscar);
	}	
	
	/**
	 *   Liste des étoiles les plus proches de l'étoile this
	 *   
	 * @param tabStars : Tableaux d'étoiles où on cherche les étoiles proches
	 * @param nmax     : Nombre maximum d'étoiles à conserver
	 * @return         : Liste des étoiles
	 */
	public ArrayList<Etoile> proxy(Etoile tabStars[], int nmax)
	{
		ArrayList<Proxy> prox = new ArrayList<Proxy>();
		int nprox = 0;
		double distmin = 1.e+10, dist;
		for (int i = 0; i < tabStars.length; i++)
		{
			double dA, dB, bM;
			Etoile e = tabStars[i];
			if (!e.flCoher) continue;
			dB = de0_dd - e.de0_dd;
			bM = (de0_dd + e.de0_dd)/2.;
			dA = (ad0_hd - e.ad0_hd)* Math.cos(Math.toRadians(bM))*15.;
			dist = Math.sqrt(dB*dB + dA*dA);
			if (nprox < nmax || dist < distmin)
			{
				distmin = dist;
				if (nprox < nmax) {nprox++; prox.add(new Proxy(dist, i));}
				else
				{
					Collections.sort(prox);
					prox.remove(nmax-1);
					prox.add(0, new Proxy(dist, i));
				}
			}			
		}
		Collections.sort(prox);
		ArrayList<Etoile> starProx = new ArrayList<Etoile>();

		for (int i = 0; i < prox.size(); i++)
		{
			Etoile e = tabStars[prox.get(i).n];
			e.work = prox.get(i).dist;
			starProx.add(e);
		}			
		return starProx;	
	}
	
	/*
	 * Module de la vitesse de l'étoile, en seconde de degré par annnée
	 */
	public double vitMod_sda()
	{
		double vX_sda = vA_sda * Math.cos(Math.toRadians(de0_dd));
		return Math.sqrt(vX_sda*vX_sda + vD_sda*vD_sda);
	}
	
	/**
	 *   Lecture du tableau des 9110 étoiles du BrightStar Catalog
	 *   La lecture est effectuée dans un stream, ce qui permet de lire
	 *   dans un fichier ressource.
	 *   
	 * @param isr  : InputStreamReader du fichier des étoiles à lire
	 * @return     : le tableau des étoiles lues.
	 */
	/*
	public static Etoile[] lect9110Stars(InputStreamReader isr)
	{
		final int MAXSTARS = 9110;
		Etoile tabStars[] = new Etoile[MAXSTARS+1];
		int nbs = 0;
		tabStars[0] = new Etoile(); tabStars[0].flCoher = false;// On commence à 1
		try {
			BufferedReader in = new BufferedReader(isr);
			byte []line;
			while(nbs < MAXSTARS)
			{
				nbs += 1;
				Etoile e = new Etoile();
				tabStars[nbs] = e;
				try {
					String str = in.readLine(); 				
					if (str == null) break; // break si fin de fichier
					line = str.getBytes();				
					if (line.length < 107) continue; // break si lignes incompletes
					e.bsNumber = Integer.parseInt(new String(line,0,4).trim());
					e.id = new String(line,4,10).trim();
					if (e.id.length() == 0)
						e.id = new String(line,51,9).trim();
					// Lignes bizarres, on ne teste que l'ascension droite)
					if(line[76]==' ' || line[78]==' ' || line[80]==' '){ e.flCoher = false; continue;}
					int HH = Integer.parseInt(new String(line,75,2));
					int MM = Integer.parseInt(new String(line,77,2));
					double SS = Double.parseDouble(new String(line,79,4));
					e.ad0_hd = HH + MM/60. + SS/3600.;
					int DD = Integer.parseInt(new String(line,84,2));
					int MD = Integer.parseInt(new String(line,86,2));
					int SD = Integer.parseInt(new String(line,88,2));
					e.de0_dd = DD + MD/60. + SD/3600.;
					if (line[83]=='-') e.de0_dd = -e.de0_dd;
					e.tipe = new String(line,127,20).trim();
					double cosDec = Math.cos(Math.toRadians(e.de0_dd));
					double vX_sda = Double.parseDouble(new String(line,148,6));
					e.vD_sda = Double.parseDouble(new String(line,154,6));
					e.vA_sda = vX_sda/cosDec;
//					e.vitMod_sda = Math.sqrt(vX_sda*vX_sda + e.vD_sda*e.vD_sda);
					e.mag = Double.parseDouble(new String(line,102,5));
					e.flCoher = true;
				}
				catch(Exception exc)
				{
					e.flCoher = false;
					System.out.println("Erreur lecture fichier 9110Stars ligne " + nbs + "; " + exc.getMessage());
					//	Toast.makeText(this, "Erreur lecture fichier " + exc.getMessage(), Toast.LENGTH_LONG).show(); return;
				}
			}	
			in.close();
		}
		catch(Exception exc)
		{
			System.out.println("Erreur lecture fichier 9110Stars ligne " + nbs);
			System.out.println(exc.getMessage());
			exc.printStackTrace();
			// Toast.makeText(this, "Erreur lecture fichier " + exc.getMessage(), Toast.LENGTH_LONG).show(); return;
		}
		return tabStars;
	}
	*/

	/**
	 *  Lecture du Bright Star Catalog
	 *  supposés être situés dans le répertoire courant du projet Java
	 *  Méthode utilisée uniquement en mode Java pur.
	 *  NON UTILISEE SOUS ANDROID
	 *  
	 * @return  : Le tableau lu, null si erreur
	 */
	/*
	public static Etoile [] loadBrightStarCatalog()
	{
		Etoile [] brightStars = null;
		FileInputStream fis2;
		try {
			fis2 = new FileInputStream(new File("stars9110.txt"));
			InputStreamReader isr2 = new InputStreamReader(fis2, "UTF-8");		
			brightStars = Etoile.lect9110Stars(isr2);	
			fis2.close();
		}		
		catch(Exception exc)
		{
			System.out.println(exc.getMessage());
			exc.printStackTrace();
		}
		return brightStars;
	}
    */
	/*
	 * Pour test sous java pur : NON UTILISEE SOUS ANDROID
	 */
	/*
	public static void main(String[] args) {
		Etoile brightStar[] = loadBrightStarCatalog();

		if (brightStar != null)
		{
			double ju =  2431559.7;
			Etoile pol = brightStar[424];
			Astro.update_eco(ju);
			System.out.println(pol.infos(ju, 1.5731, 43.5254, null, 1, true));
			double tHA_dg[] = {0.};
			double pis = Astro.polarisInScope_dg(ju, 1.5731, tHA_dg, true);
			System.out.println("Polaris in Scope : " + Utils.deg2hms(pis));
			System.out.println("Angle horaire    : " + Utils.deg2hms(tHA_dg[0]));

			double vmax = 0.;
			int imax = -1;
			int nok = 0;
			for (int i = 0; i < brightStar.length; i++)
			{
				Etoile e = brightStar[i];
//				if (!e.ok9110) 
//					continue;
				nok++;
				double vitMod = e.vitMod_sda();
				if (vitMod > vmax) 
				{
					vmax = vitMod; 
					imax = i;
				}
			}
			System.out.println("Il y a " + nok + " étoiles ok!");
			
			if(imax >= 0)
			{
				Etoile e = brightStar[imax];
				System.out.println("L'étoile  " + e.id + ", numero " + e.bsNumber + ", mag = "+ e.mag + ", est la plus rapide\n"
						+ "Vitesse = " + e.vitMod_sda() + " secondes de degrés par an");
			}
		}			
	}
	*/
}
