package fr.llibre.ml2;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;


/**
 *   Classe permettant de decoder les TLE
 * 
 * TLE veut dire Two-Line Elements en anglais. Ces deux lignes continennent les
 * elements orbitaux qui définissent l'orbite exacte d'un satellite et la
 * position du satellite sur son orbite au moment precis appele epoque. L'epoque
 * des donnees orbitales, ou l'heure exacte a laquelle les donnees presentees sont
 * valides, est inscrite a la ligne un des TLE. Elle sert a determiner l'age des
 * elements, donc de savoir si les previsions qui seront generees avec eux seront
 * precises.
 * Les TLE d'un grand nombre de satellites sont accessibles sur le site :
 * http://celestrak.com/NORAD/elements/
 * @author Michel Llibre
 */
public class TwoLE {
	public String name;
	int catnum; 		/* Numero identifiant le satellite */
	int setnum;			/* Numero bulletin de cette tle (raz a 9999) */
	String designator;	/* "AAxxxxxx" = xxxxx numero du lancement dans l'annee AA */
	int year;			/* Annee de la date du releve des donnees */
	double refepoch;	/* Jour fractionnaire de cette date dans l'annee */
	double incl;		/* Inclinaison de l'orbite en degres */
	double raan;		/* Righ Ascension of the Ascending Node (degres) */
	double eccn;		/* Excentricite de l'orbite */
	double argper;		/* Argument du perigee (degres) */
	double meanan;		/* Anomalie moyenne a la date des donnees (degrees) */
	double meanmo;		/* Mouvement moyen en tours/jour */
	double drag;		/* Ralentissement du mouv.moy en tours/jour/jour */
	double nddot6;		/* Derivee 3eme de la revolution en tours/jour/jour/jour */
	double bstar;		/* Terme de trainee ?? */
	int orbitnum;		/* Nombre de revolutions depuis de lancement */	
	
public final static String norad =  "http://celestrak.com/NORAD/elements/";

public final static String[] typeSats = {"stations", "tle-new", "visual", "science",
	"weather", "resource", "sarsat", "geo", "geodetic", "intelsat",	"ses",
	"iridium", "globalstar", "gps-ops", "glo-ops", "galileo", "beidou"};

public final static String[] url = urlList();

    /**
     *  Obtention d'une liste d'url contenant des tle
     * @return La liste de String des url contenant des tle
     */
    private static String[] urlList()
    {
    	String[] list = new String[typeSats.length];
    	for (int i = 0; i < typeSats.length; i++) list[i] = norad + typeSats[i] + ".txt";
    	return list;
    }
    
    /**
     *  Obtention de la liste des satellites figurant à une url 
     * @param strURL : Une adresse susceptible de contenir des TLE
     * @return La liste des satellites présents à cette adresse
     */
    public static ArrayList<String> satList(String strURL)
    {
    	ArrayList<String> liste = new ArrayList<String>();
		try {
	        URL url = new URL(strURL);
	        BufferedReader in;
			
			in = new BufferedReader(new InputStreamReader(url.openStream()));
	        String inputLine;
	        int n = 0;
	        while ((inputLine = in.readLine()) != null)
	        {
	        	n++;
	        	if (n%3 == 1)liste.add(inputLine.trim());
	        }
	        in.close();			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return liste;
    }

	/**
	 *  Obtention d'un Tableau de noms de satellite figurant à une url
	 * @param strURL : Une adresse susceptible de contenir des TLE
	 * @return Le tableau des noms de satellites présents à cette adresse
	 */
	public static String[] satArray(String strURL)
	{
		ArrayList<String> alistSats = satList(strURL);
		int n = alistSats.size();
		if (n > 0) {
			String[] listSats = new String[n];
			for (int i = 0; i < n; i++) listSats[i] = alistSats.get(i);
			return listSats;
		}
		return null;
	}

    
    /**
     *  Construction d'une TLE à partir d'une url et du nom du satellite
     * @param strURL : Une adresse susceptible de contenir des TLE
     * @param nom : Nom d'un satellite dans la liste des TLE
     * @return
     */
    public static TwoLE makeNew(String strURL, String nom)
    {
    	TwoLE tle = null;
		try {
	        URL url = new URL(strURL);
	        BufferedReader in;
			
			in = new BufferedReader(new InputStreamReader(url.openStream()));
	        String inputLine, str1, str2;
	        int n = 0;
	        while ((inputLine = in.readLine()) != null)
	        {
	        	n++;
	        	if (n%3 != 1) continue;
	        	if (!inputLine.trim().equalsIgnoreCase(nom)) continue;	
	        	str1 = in.readLine(); if (str1 == null) break;
	        	str2 = in.readLine(); if (str2 == null) break;
	        	byte [] line1 = str1.getBytes();
	        	byte [] line2 = str2.getBytes();
				int res = check(line1,line2);
				if (res < 0)
				{
					System.out.println("Erreur checksum TLE satellite " + nom);
					break;
				}
				if (res > 0)
				{
					System.out.println("Erreur codage TLE satellite " + nom);
					break;
				}
				tle = new TwoLE(nom, line1, line2);
				break;	        	
	        }
	        in.close();			
		} catch (IOException e) {e.printStackTrace();}
		return tle;
    }
    
       
    /**
     *  Construction d'une TLE à partir d'une url et du rang d'un satellite
     * @param strURL : Une adresse susceptible de contenir des TLE
     * @param rang : Rang d'un satellite dans le fichier des TLE (0 à n-1)
     * @return
     */
    public static TwoLE makeNew(String strURL, int rang)
    {
    	TwoLE tle = null;
		try {
	        URL url = new URL(strURL);
	        BufferedReader in;
			
			in = new BufferedReader(new InputStreamReader(url.openStream()));
	        String inputLine, str1, str2;
	        int n = 0;
	        while ((inputLine = in.readLine()) != null)
	        {
	        	if (n++ != rang*3) continue;	
	        	String nom = inputLine.trim();
	        	str1 = in.readLine(); if (str1 == null) break;
	        	str2 = in.readLine(); if (str2 == null) break;
	        	byte [] line1 = str1.getBytes();
	        	byte [] line2 = str2.getBytes();
				int res = check(line1,line2);
				if (res < 0)
				{
					System.out.println("Erreur checksum TLE satellite rang " + rang);
					break;
				}
				if (res > 0)
				{
					System.out.println("Erreur codage TLE satellite rang " + rang);
					break;
				}
				tle = new TwoLE(nom, line1, line2);
				break;	        	
	        }
	        in.close();			
		} catch (IOException e) {e.printStackTrace();}
		return tle;
    }
    
	/**
	 *  Construction et decodage d'une TLE
	 * @param name		String nom du satellite
	 * @param line1		Ligne 1
	 * @param line2		Ligne 2
	 */
    private TwoLE(String nom, byte []line1, byte []line2)
    {
		double tempnum;
		name = new String(nom);
		designator = new String(line1,9,8);
		catnum = Integer.parseInt((new String(line1,2,5)).trim());
		year = Integer.parseInt(new String(line1,18,2));
		if (year <  56) year += 2000; else year += 1900;
		refepoch = Double.parseDouble(new String(line1,20,12));
		drag = Double.parseDouble(new String(line1,33,10));
		tempnum = 1.0e-5* Double.parseDouble(new String(line1,44,6));
		if(tempnum!= 0.) nddot6 = tempnum/ Math.pow(10.0,(line1[51]-'0'));
		else nddot6 = 0.;
		tempnum = 1.0e-5* Double.parseDouble(new String(line1,53,6));
		if(tempnum!= 0.) bstar = tempnum/ Math.pow(10.0,(line1[60]-'0'));
		else bstar = 0.;
		setnum = Integer.parseInt((new String(line1,64,4)).trim());
		incl = Double.parseDouble(new String(line2,8,8));
		raan = Double.parseDouble(new String(line2,17,8));
		eccn = 1.0e-07* Double.parseDouble(new String(line2,26,7));
		argper = Double.parseDouble(new String(line2,34,8));
		meanan = Double.parseDouble(new String(line2,43,8));
		meanmo = Double.parseDouble(new String(line2,52,11));
		orbitnum = Integer.parseInt((new String(line2,63,5)).trim());
    }
    
	/**
	 * Verification du format des TLE
	 * @param line1
	 * @param line2
	 * @return 0 si OK, -1 si erreur Checksum, 1 si autre erreur
	 */
	private static int check(byte []line1, byte []line2)
	{
		if (line1.length < 69 || line2.length < 69) return 1;	
		int x;
		int sum1, sum2;
		// initialisation val
		int i;
		byte []val = new byte[256];
		for (i = 0  ; i <= 255; val[i] = 0, i++);
		for (i = '0'; i <= '9'; val[i] = (byte) (i - '0'), i++);
		val['-'] = 1;		
        // Verification TLE
		for (x=0, sum1=0, sum2=0; x<=67; 
		sum1 +=  val[(int)line1[x]], 
		sum2 +=  val[(int)line2[x]], x++);

		x = ( val[(int)line1[68]]^(sum1%10));
		x |= (val[(int)line2[68]]^(sum2%10));
		if (x != 0) return -1;
		x = (line1[0]^'1') ;
		x |= (line1[1]^' ') ;
		x |= (line1[7]^'U') ;
		x |= (line1[8]^' ') ;
		x |= (line1[17]^' ') ;
		x |= (line1[23]^'.') ;
		x |= (line1[32]^' ') ;
		x |= (line1[34]^'.') ;
		x |= (line1[43]^' ') ;
		x |= (line1[52]^' ') ;
		x |= (line1[61]^' ') ;
		x |= (line1[62]^'0') ;
		x |= (line1[63]^' ') ; 
		x |= (line2[0]^'2')  ;
		x |= (line2[1]^' ')  ;
		x |= (line2[7]^' ')  ;
		x |= (line2[11]^'.') ;
		x |= (line2[16]^' ') ;
		x |= (line2[20]^'.') ;
		x |= (line2[25]^' ') ;
		x |= (line2[33]^' ') ;
		x |= (line2[37]^'.') ;
		x |= (line2[42]^' ') ;
		x |= (line2[46]^'.') ;
		x |= (line2[51]^' ') ;
		x |= (line2[54]^'.') ;
		x |= (line1[2]^line2[2]) ;
		x |= (line1[3]^line2[3]) ;
		x |= (line1[4]^line2[4]) ;
		x |= (line1[5]^line2[5]) ;
		x |= (line1[6]^line2[6]) ;
		x |= (Character.isDigit(line1[68]) ? 0 : 1) ;
		x |= (Character.isDigit(line2[68]) ? 0 : 1) ;
		x |= (Character.isDigit(line1[18]) ? 0 : 1) ;
		x |= (Character.isDigit(line1[19]) ? 0 : 1) ;
		x |= (Character.isDigit(line2[31]) ? 0 : 1) ;
		x |= (Character.isDigit(line2[32]) ? 0 : 1) ;
		return (x != 0 ? 1 : 0);
	}	
    
	/**
	 * Impression d'une TwoLE
	 */	
	public String infos()
	{
		String texte =
		"Nom du satellite :                " + name +
		"\nNuméro dans le Catalogue :        " + catnum +
		"\nElement set :                     " + setnum +
		"\nDésignation :                     " + designator +
		"\nAnnée de la date des données :    " + year +
		"\nJour dans l'année :               " + refepoch +
		"\nInclinaison orbite :              " + incl + " degres" +
		"\nAscension droite Noeud.Asc. :     " + raan + " degres" +
		"\nExcentricité :                    " + eccn +
		"\nArgument du perigée :             " + argper + " degres" +
		"\nAnomalie moyenne à la date :      " + meanan + " degres" +
		"\nMouvement moyen :                 " + meanmo + " tours/jour" +
		"\n1/2 der1 mouv. moy :              " + drag + " tours/jour**2" +
		"\n1/6 der2 mouv. moy :              " + nddot6 + " tours/jour**3" +
		"\nTrainée bstar :                   " + bstar + " ?" +
		"\nNombre de révolutions à la date : " + orbitnum;
		return texte;
	}
	
    public static void main(String[] args) throws Exception
    {
		for (int i = 0; i < TwoLE.url.length; i++) System.out.println(i + " : " + TwoLE.url[i]);
		if (TwoLE.url.length <= 0) return;

		// On prend la 1ere : les tle des tations et on construit la liste des noms
		ArrayList<String> listSats = TwoLE.satList(TwoLE.url[0]);
		if (listSats.size() <= 0) return;
		for (int i = 0; i < listSats.size(); i++) System.out.println(i + " : " + listSats.get(i));

		// On le 1er nom dans la liste (l'ISS Zarya)
		TwoLE tle = TwoLE.makeNew(TwoLE.url[0], listSats.get(0));
		System.out.println(tle.infos());
    }    
}
